#ifndef __SMARTLOW_H__
#define __SMARTLOW_H__

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <sys/time.h>
#include <signal.h>
#include <ctype.h>
#include <stdio.h>

/** Define the maximum length for a command. */
#define MAX_CMD_LENGTH 100

/** 
 * Define the default interval assumed for the internal timing .
 * It is expressed in micro-seconds.
 */
#define INTERVAL 100000

/** Timeout in seconds for waiting a message from the server */
#define MAX_WAIT_PERCEPT 5

/** Size of the incoming message's buffer. */
#define MSG_BUF_SIZE 2048

/** 
 * Coefficient used to calculate the new timer value. The 
 * classical formula is 
 * 
 * (alpha * old_val) + ((1 - alpha) * new_val)
 *
 */
#define ALPHA 0.9

#include "parser.h"
#include "queue.h"

/** Constants used for indicate the perception type */
typedef enum {
  SMPT_ERROR,
  SMPT_INIT,
  SMPT_SEE,
  SMPT_HEAR,
  SMPT_REFEREE,
  SMPT_SENSE_BODY
} smart_perception_type;

/** Definition of the sighandler type for the setitimer system call */
typedef void (*sighandler_t)(int);

/** 
 * Constants used for indicate the modality used 
 * for sending commands.
 */
typedef enum {
  SEND_QUEUE,
  SEND_NOW
} smart_send_mode;

/** Structure that describes a socket */
typedef struct {
    int			socketfd;
    struct sockaddr_in	serv_addr;
} smart_socket;

/** 
 * Structure that describes
 * sense_body message.
 */
typedef struct {
  int time;
  rc_view_quality_t quality;
  rc_view_width_t viewwidth;
  int stamina;
  int effort;
  float speed;
  float headdir;
  int kickcount;
  int dashcount;
  int turncount;
  int saycount;
  int turnneckcount;
} smart_sense_body;

/**
 * Global variables
 */

/** Array that holds the labels "narrow" "normal" and "wide" */
extern char* angle_width_v[];

/** Array that holds the labels "high" and "low" */
extern char* quality_v[];

/** This variable contains the last sense_body received */
extern smart_sense_body percept_sense_body;

/** Type of the last perception */
extern smart_perception_type perception_type;

/** Boolean variable that say if turn trashing is on or off */
extern int turn_trash;

/** Buffer used for sending commands */
extern char msgbuf[];

/**
 * Public functions
 */

/**
 * Open the connection with the server soccer.
 * @param host Hostname or IP address of the server
 *             where the soccerserver is running.
 * @param port Port used for the connection (usually 6000)
 * @return 0 is returned if the connection is ok -1 otherwise.
 */
int smart_connect_server_soccer(const char* host, const int port);

/**
 * Close the connection with server.
 */
void smart_close_connection();

/**
 * Read a message from the server. This is a synchronous
 * call. She waits for the message.
 * @return A pointer to the buffer that contains the message.
 */
char* smart_read_percept();

/**
 * Read the last message from the server. Other messages
 * are discarded. If a (hear) message is found the message
 * is returned and the function ends. In this case maybe that
 * other messages are leaved on the socket queue. The rationale
 * behind this semantics is that (hear) messages may contain
 * referee messages that represent the state of the game.
 * @return A pointer to the buffer that contains the message.
 */
char* smart_read_last_percept();

/**
 * Send a command to the server. The second parameter specifies
 * the mode of transmission: using the SEND_NOW constant the
 * message is delivered immediatly; otherwise the SEND_QUEUE
 * constant indicates that the message must be enqueued and
 * sent when the timer expires.
 * The preffered mode is the second one because the timer
 * guarantees that a command is sent using the optimal timing
 * without loss of commands. For this reason the macro defined
 * for sending commands use this mode. The only exception is
 * the macro for sending (turn-neck) command because this command
 * is received by the server also if coupled with other commands.
 * @param cmd Command that must be sent to the server.
 * @param mode If it is SEND_NOW the command is immediatly sent;
 *             otherwise if it is SEND_QUEUE the command is added
 *             to the command's queue.
 */
void smart_send_command(char* cmd, smart_send_mode mode);

/**
 * Change the value for the ALPHA parameter.
 * @param a New value for alpha
 * @return The old value for the parameter.
 * @see ALPHA
 */
int smart_set_alpha_timer(float a);

/**
 * Change the value for the turn_trash flag. When turn trashing is on
 * the commands' queue becomes empty before the insertion of a 'turn'
 * command with a moment greater than 10. The rationale for this behavior
 * is that turning changes substantially the world perceived by the
 * player and thus, if the internal state delivers a turn action means
 * that the expectation is that the action is performed immediately.
 * The insertion of a dash command in queue does not change too much
 * the world perceived and then the queue is preserved.
 * The turn trashing can be disable because sometime maybe useful
 * sending series of commands (i.e. for turning with ball) that
 * contains turn actions interleaved with other actions.
 * @param tt It can be 1 or 0 to enable or disable the turn trashing.
 * @return The previous value of the flag.
 */
int smart_set_turn_trash(int tt);

/**
 * Returns 1 if turn trashing is enable 0 otherwise.
 * @return The value of turn trashing flag.
 * @see smart_set_turn_trash(int)
 */
int smart_get_turn_trash();

/**
 * Check if a message from the server is available without
 * parsing the new message. To get the message use smart_read_percept()
 * or smart_read_last_percept(). This call is non blocking.
 * @return 1 if the message is available 0 otherwise.
 * @see smart_read_percept()
 * @see smart_read_last_percept()
 */
int smart_is_percept_available();

/**
 * Send the (init) command to the soccerserver.
 * @param ISGOALIE  0 is a normal player ow is a goalie
 * @param TEAMNAME  String that holds the team's name
 * @param VERSIONNO Version of the protocol (is a float value).
 */
#define smart_send_init(ISGOALIE, TEAMNAME, VERSIONNO) \
  { \
    char *goalie = (ISGOALIE)?" (goalie)":""; \
    sprintf(msgbuf, "(init %s%s (version %f))", TEAMNAME, goalie, VERSIONNO); \
    smart_send_command(msgbuf, SEND_NOW); \
  }

/**
 * Send the (turn) command to the soccer server.
 * @param DIR Turn moment. Must be an integer within [-180, 180].
 */
#define smart_send_turn(DIR) \
  { \
    sprintf(msgbuf, "(turn %i)", DIR); \
    smart_send_command(msgbuf, SEND_QUEUE); \
  }

/**
 * Send the (turn-neck) command to the soccer server.
 * This command is sent immediatly.
 * @param ANGLE Angle of the neck. Must be an integer within [-90, 90].
 */
#define smart_send_turn_neck(ANGLE) \
  { \
    sprintf(msgbuf, "(turn-neck %i)", ANGLE); \
    smart_send_command(msgbuf, SEND_NOW); \
  }

/**
 * Send the (dash) command to the soccer server.
 * @param POW Speed used for run. Must be an integer within [-30, 100].
 */
#define smart_send_dash(POW) \
  { \
    sprintf(msgbuf, "(dash %i)", POW); \
    smart_send_command(msgbuf, SEND_QUEUE); \
  }

/**
 * Send the (catch) command to the soccer server.
 * @param DIR Turn moment. Must be an integer within [-180, 180].
 */
#define smart_send_catch(DIR) \
  { \
    sprintf(msgbuf, "(catch %i)", DIR); \
    smart_send_command(msgbuf, SEND_QUEUE); \
  }

/**
 * Send the (say) command to the soccer server.
 * @param MSG String that contains the message that must be sent.
 */
#define smart_send_say(MSG) \
  { \
    sprintf(msgbuf, "(say %s)", MSG); \
    smart_send_command(msgbuf, SEND_QUEUE); \
  }

/**
 * Send the (kick) command to the soccer server.
 * @param POW Integer that indicates the power of the kick [-30, 100).
 * @param DIR Integer that indicates the direction of the kick [-180, 180].
 */
#define smart_send_kick(POW, DIR) \
  { \
    sprintf(msgbuf, "(kick %i %i)", POW, DIR); \
    smart_send_command(msgbuf, SEND_QUEUE); \
  }

/**
 * Send the (change_view) command to the soccer server.
 * @param AW Angle width (use one of rc_view_width_t constants
 *           defined in the file parser.h)
 * @param QY Quality view (use one of rc_view_quality_t constants
 *           defined in the file parser.h)
 */
#define smart_send_change_view(AW, QY) \
  { \
    sprintf(msgbuf, "(change_view %s %s)", angle_width_v[AW], quality_v[QY]); \
    smart_send_command(msgbuf, SEND_QUEUE); \
  }

/**
 * Send the (move) command to the soccer server.
 * @param X Integer that specifies the x-coordinate
 * @param Y Integer that specifies the y-coordinate
 */
#define smart_send_move(X, Y) \
  { \
    sprintf(msgbuf, "(move %i %i)", X, Y); \
    smart_send_command(msgbuf, SEND_QUEUE); \
  }

#endif /* __SMARTLOW_H__ */
