;;;; agent-kb.lsp
;;;; 
;;;; Last update: 10/08/1999
;;;; @author Antonio Cisternino
;;;; @version 0.5
;;;;
;;;; Purpose: This file define the KB interface of the agent
;;;;
;;;; The directions of a player are:
;;;;
;;;;            NW N NE
;;;;             W P E
;;;;            SW S SE
;;;; 
;;;; The position of the player is expressed in 1/16th of field
;;;;
;;;; Changes:
;;;;
;;;; 10/08/1999: Now CLOS is not used
;;;;
;;;; 11/4/1999: Fixed a bug: the switch-side  must not invoked when
;;;;            half_time massage occours.
;;;;
;;;; 13/3/1999: Added the goalie-fire slot
;;;;
;;;; 6/3/1999: Added the method change-dir
;;;;
;;;; 9/1/1999: The enemy-goal is now a flat list:
;;;;                (TIME DIST DIR ...)
;;;;
;;;; 9/1/1999: Now the state game offside is handled correctly
;;;;
;;;; 9/1/1999: Now sense body is implemented with the neck extensions
;;;;           (server 5.XX). There is a little irregularity: the 
;;;;           is stored in the pos-info slot rather than in the
;;;;           sense-body slot.
;;;;
;;;; 9/1/1999: Constants have been  moved to constants.lsp, macros have been
;;;;           moved to the file macros.lsp. Definition of structures
;;;;           have been moved to the file struct.lsp.
;;;;
;;;; 9/1/1999: Fixed a bug: the type referee for a message was never thrown
;;;;
;;;; 9/1/1999: Fixed a bug: the last-message-read flag was never reset
;;;;
;;;; 9/1/1999: Added the support for the sense-body perception. The
;;;;           KB has been modified and a sense-body structure
;;;;           has been defined.
;;;;           *NOTE*: The fundamental hypotesys is that the
;;;;           sense-body perception is sent automatically like
;;;;           in the server 5.XX. Thus there are no default values
;;;;           assuming that the information is obtained quickly
;;;;           and is available when needed by various predicates
;;;;
;;;; 9/1/1999: Fixed a serious bug: in the cond that decide for the
;;;;           type of referee message GOAL_ was before GOAL_KICK and
;;;;           thus the matcher will fails.
;;;;
;;;; 9/1/1999: obj-near and obj-very-near have been modified and now
;;;;           works properly
;;;;
;;;; 8/1/1999: Random values are choosed at run-time. obj-near and
;;;;           obj-very near are now defined without random.
;;;;
;;;; 6/1/1999: Added the macro obj-near, obj-very-near. Changed the
;;;;           macros without parameters: now are calculed before
;;;;           the expansion
;;;;
;;;; 3/1/1999: Added the last-message-read flag
;;;;
;;;; 2/1/1999: In the referee section I've changed "CORNER-KICK_" with
;;;;           "CORNER_KICK_". It seems a bug but I must check the manual
;;;;
;;;; 2/1/1999: added a lot of comments

;;; Knowledge Base of the agent
(defstruct agent-kb
  ;; Last percept about my team's goal
  my-goal

  ;; Last percept about the enemy team's goal
  ;; Contains a list of type:
  ;; 
  ;;   (TIME DIST DIR ...)
  ;;
  ;; The list obtained prepending the time to the cdr of the
  ;; list (GOAL ...) of the object goal contained in the perception
  enemy-goal

  ;; Represent the position of the ball
  ball-pos

  ;; Position of the other players of my team. To use the number
  ;; of the shirt the size of the array is 12.
  (team-players (make-array 12 :initial-element nil))

  ;; Position of the players of the oter team. To use the number
  ;; of the shirt the size of the array is 12.
  (enemy-players (make-array 12 :initial-element nil))
   
  ;; Last perception stored in the KB
  last-percept

  ;; Type of the last perception: can be 'SEE, 'MESSAGE, 'REFEREE,
  ;; 'SENSE-BODY or 'ERROR.
  last-percept-type

  ;; Contains the last message of the referee. In fact contains the
  ;; state of the game
  last-referee-msg 

  ;; last-message-read
  last-message-read

  ;; Name of the team
  team
   
  ;; Last message heard
  last-msg

  ;; Contains the current side of the team.
  side

  ;; evt??? I don't know why this attribute exists.
  msg-buf

  ;; Position of the player
  player-pos

  ;; Status of the player
  (sense-body-info (make-sense-body))

  ;; Throw ball?
  goalie-fire
  
  ;; For the moment is unused
  history

  ;; Seen ball dir
  last-ball-dir

  ;; Seen ball dist
  last-ball-dist

  ;; Front players
  (front-players 0)

  ;; Left players
  left-players

  ;; Right players
  right-players)

(defun switch-side (kb)
  "Changes the value of the side switch in the agent KB"
  (setf (agent-kb-side kb)
	(if (char= (agent-kb-side kb) #\L)
	    #\R
	  #\L)))

(defun message-read (kb)
  "Set to read the last-message-read flag"
  (setf (agent-kb-last-message-read kb) t))

(defun change-dir (kb moment)
  "Changes the direction of the player assuming that
it turns by 'moment' degrees."
  (let* ((p (agent-kb-player-pos kb))
	 (dir (+ (pos-dir p) moment)))
    (when (> (abs dir) 180)
	(if (> dir 0)
	    (incf dir -360)
	  (incf dir 360)))
    (setf (pos-dir p) dir))
  nil)

(defun tell-KB (kb p)
  "Adds a perception to the KB"

  ;; update the value of the last-percept attribute
  (setf (agent-kb-last-percept kb) p)

  ;; check for the type of the percept
  (if (eq 'see (car p)) ; SEE percept
      (progn
	;; calculate the absolute player position
	(calc-player-pos kb p)

	;; mark contains the simulation step (the time)
	(let (obj-name num (mark (second p)))

	  ;; update the type of the last percept
	  (setf (agent-kb-last-percept-type kb) 'see)

	  ;; Direction and distance of ball not known
	  (setf (agent-kb-last-ball-dir kb) nil
		(agent-kb-last-ball-dist kb) nil)

	  ;; Obstacles are not known
	  (setf (agent-kb-front-players kb) 0
		(agent-kb-left-players kb) 0
		(agent-kb-right-players kb) 0)

	  ;; iterate foreach object and update the KB
	  (dolist 
	   (obj (cddr p))
	   
	   ;; Take the name of the object and handle the various cases
	   (setf obj-name (caar obj))

	   (cond

	    ;; (GOAL ...)
	    ((eq 'goal obj-name)
	     ;; update the goal information. The information is
	     ;; useful only if the current time is equal to mark
	     (if (char= (agent-kb-side kb) (if (eq 'l (cadar obj)) #\L #\R))
		 (setf (agent-kb-my-goal kb) `(,mark ,@(cdr obj)))
	       (setf (agent-kb-enemy-goal kb) `(,mark ,@(cdr obj)))))

	    ;; (PLAYER ...)
	    ((eq 'player obj-name)

	     ;; Update obstacles information

	     ;; Angle limit is calculated using the
	     ;; function y = a/x +c and f(1) = 45, f(20) = 0
	     ;; a is (20 /19) * 45 and c is - (45 / 19)
	     ;; x is the player distance
	     ;; The second angle limit is calculated in the
	     ;; same way. Conditions are f(1) = 90, f(50) = 0 that
	     ;; implies a = (50 / 49) * 90 and c = - (90 / 49)
	     (let ((angle-limit (- (/ 47.4 (second obj)) 2.4))
		   (angle-limit-2 (- (/ 91.8 (second obj)) 1.8)))
	       (if (and (> angle-limit 0)
			(< (abs (third obj)) angle-limit))
		   (incf (agent-kb-front-players kb))
		 (when (and (> angle-limit-2 0)
			    (< (abs (third obj)) angle-limit-2))
		   (if (> (third obj) 0)
		       (incf (agent-kb-right-players kb))
		     (incf (agent-kb-left-players kb))))))

	     ;; extract the player number
	     (setf num (caddar obj))

	     ;; If the percept is enough detailed we can store
	     ;; the information in the KB.
	     ;; Note that this is also a filter: if a player is
	     ;; far in general is not relevant.
	     (when (and num (cadar obj))

	       ;; is a player of my team
	       (if (eq (agent-kb-team kb) (cadar obj))
		   (progn
		     ;; if necessary allocate a structure
		     (unless (svref (agent-kb-team-players kb) num)
		       (setf (svref (agent-kb-team-players kb) num) (make-obj-info)))
		     
		     ;; Store the information calculating the absolutes
		     ;; coordinates
		     (setf (obj-info-time (svref (agent-kb-team-players kb) num)) mark)
		     (calc-obj-pos kb obj (svref (agent-kb-team-players kb) num)))

		 ;; is a player of the opposite team
		 (progn
		   ;; if necessary allocate a structure
		   (unless (svref (agent-kb-enemy-players kb) num)
		     (setf (svref (agent-kb-enemy-players kb) num) (make-obj-info)))

		   ;; Store the information calculating the absolutes
		   ;; coordinates
		   (setf (obj-info-time (svref (agent-kb-enemy-players kb) num)) mark)
		   (calc-obj-pos kb obj (svref (agent-kb-enemy-players kb) num))))))
	    
	    ;; (BALL ...)
	    ((eq 'ball obj-name)

	     ;; Direction and distance of ball not known
	     (setf (agent-kb-last-ball-dir kb) (third obj)
		   (agent-kb-last-ball-dist kb) (second obj))

	     ;; if necessary allocate a structure
             (unless (agent-kb-ball-pos kb)
	       (setf (agent-kb-ball-pos kb) (make-obj-info)))

	     ;; Store the information calculating the absolutes
	     ;; coordinates
	     (setf (obj-info-time (agent-kb-ball-pos kb)) mark)
	     (calc-obj-pos kb obj (agent-kb-ball-pos kb)))

	    ;; Otherwise
	    (t nil)))))
    
    ;; Else for the type of the precept
    ;; hear perception
    (if (eq 'hear (car p)) ; HEAR percept

      ;; extract the various fields from the percept
      (let ((mark (second p)) (dir (third p)) (msg (fourth p)))

	;; if the direction is REFEREE handle the messages
	(if (eq dir 'referee)
	    (let ((msgs (princ-to-string msg)))

	      ;; update the type of the last percept
	      (setf (agent-kb-last-percept-type kb) 'referee
		    (agent-kb-last-message-read kb) nil)

	      ;; set the state of the game
	      (setf (agent-kb-last-referee-msg kb)
		    (cond
		     ;; before_kick_off
		     ((eq 'before_kick_off msg)
		      '(before-kick-off))

		     ;; play_on
		     ((eq 'play_on msg)
		      '(play-on))

		     ;; half_time
		     ((eq 'half_time msg)
		      ;; update the state of the side switch
;		      (switch-side kb)
		      '(half-time))

		     ;; time_up
		     ;; probably here I must quit the ecl
		     ((eq 'time_up msg)
		      (switch-side kb)
		      '(time-up))
		     
		     ;; time_over
		     ((eq 'time_over msg)
		      (switch-side kb)
		      '(time-over))
		     
		     ;; extend
		     ((eq 'extend msg)
		      '(extend))

		     ;; drop_ball
		     ((eq 'drop_ball msg)
		      '(drop-ball))

		     ;; foul_...
		     ((string= msgs "FOUL_" :end1 5 :end2 5)
		      (if (char= (char msgs 5) (agent-kb-side kb))
			  '(foul my)
			'(foul enemy)))

		     ;; goal_... The and allow to distinguish between
		     ;; GOAL_ and GOAL_KICK_
		     ((and
		       (not (char= (char msgs 5) #\K))
		       (string= msgs "GOAL_" :end1 5 :end2 5))
		      (if (char= (char msgs 5) (agent-kb-side kb))
			  `(goal my ,(subseq msgs 7))
			`(goal enemy ,(subseq msgs 7))))

		     ;; offside_...
		     ((string= msgs "OFFSIDE_" :end1 8 :end2 8)
		      (if (char= (char msgs 8) (agent-kb-side kb))
			  '(kick-in my)
			'(kick-in enemy)))

		     ;; kick_in_...
		     ((string= msgs "KICK_IN_" :end1 8 :end2 8)
		      (if (char= (char msgs 8) (agent-kb-side kb))
			  '(kick-in my)
			'(kick-in enemy)))

		     ;; kick_off_...
		     ((string= msgs "KICK_OFF_" :end1 9 :end2 9)
		      (if (char= (char msgs 9) (agent-kb-side kb))
			  '(kick-off my)
			'(kick-off enemy)))

		     ;; goal_kick_...
		     ((string= msgs "GOAL_KICK_" :end1 10 :end2 10)
		      (if (char= (char msgs 10) (agent-kb-side kb))
			  '(goal-kick my)
			'(goal-kick enemy)))

		     ;; free_kick_...
		     ((string= msgs "FREE_KICK_" :end1 10 :end2 10)
		      (if (char= (char msgs 10) (agent-kb-side kb))
			  '(free-kick my)
			'(free-kick enemy)))

		     ;; corcker_kick_...
		     ((string= msgs "CORNER_KICK_" :end1 12 :end2 12)
		      (if (char= (char msgs 12) (agent-kb-side kb))
			  '(corner-kick my)
			'(corner-kick enemy)))

;		     ((string= msgs "FOUL_" :end1 5 :end2 5)
;		      (if (char= (char msgs 5) (side kb))
;			  `(foul my ,(subseq msgs 7))
;			`(foul enemy ,(subseq msgs 7))))

		     ;; Otherwise
		     (t nil)
		   )))
	  ;; Else if it is not a message that I've sent...
	  (unless (eq dir 'self)
	    ;; update the type of the message
	    (setf (agent-kb-last-percept-type kb) 'message)

	    ;; update the content of the message in the KB
	    (setf (agent-kb-last-msg kb) (cdddr p))
	    )))
      
      ;; Else is a sense_body perception
      (if (eq 'sense_body (car p)) ; SENSE_BODY percept
	  (when (< (rc-trashing) 0.5)
	    (let ((sb (agent-kb-sense-body-info kb)))
	      ;; update the type of the last percept
	      (setf (agent-kb-last-percept-type kb) 'sense-body)

	      ;; Get the time of the percept
	      (setf (sense-body-time sb) (second p))
	  
	      ;; Iterates for all the values:
	      (dolist 
	       (el (cddr p))
	       (cond

		;; VIEW_MODE field
		((eq (car el) 'view_mode)
		 (setf (sense-body-view-mode-qualty sb) (second el)
		       (sense-body-view-mode-width sb)  (third el)))

		;; STAMINA field
		((eq (car el) 'stamina)
		 (setf (sense-body-stamina sb) (second el)
		       (sense-body-effort sb)  (third el)))

		;; SPEED field
		((eq (car el) 'speed)
		 (setf (sense-body-speed sb) (second el)))

		;; KICK field
		((eq (car el) 'kick)
		 (setf (sense-body-kick-count sb) (second el)))

		;; DASH field
		((eq (car el) 'dash)
		 (setf (sense-body-dash-count sb) (second el)))
	    
		;; TURN field
		((eq (car el) 'turn)
		 (setf (sense-body-turn-count sb) (second el)))

		;; SAY field
		((eq (car el) 'say)
		 (setf (sense-body-say-count sb) (second el)))

		;; TURN_NECK field
		((eq (car el) 'turn_neck)
		 (setf (sense-body-turn-neck-count sb) (second el)))

		;; HEAD_ANGLE field
		((eq (car el) 'head_angle)
		 (setf (pos-head-angle (agent-kb-player-pos kb)) (second el)))

	       (t nil)))))

	;; Else the type of the message is unknown: error!
	(progn
	  ;; Set the type of the last percept
	  (setf (agent-kb-last-percept-type kb) 'ERROR)
	
	  (format t "Error: the server has sent the percept: ~A~%" p))))))

