;;;; 
;;;; Last update: 25/1/1999
;;;; @author Antonio Cisternino
;;;; @version 0.1
;;;;
;;;; Purpose: here there are the functions availables
;;;;          for writing plans
;;;;
;;;; Changes:
;;;;
;;;; 17/7/1999: Bugfix: the dir-ball function was a little buggy ;)
;;;;
;;;; 11/4/1999: Added the enemies-near-ball and friends-near-ball.
;;;;
;;;; 25/2/1999: Bug-fix in dir-x-y: completely rewritten
;;;;
;;;; 14/1/1999: Bug-fix in dir-x-y: atan was used instead deg-atan


(defun referee (kb)
 "Gets the last referee message (is the state of the game)
  Returns a list where the (car) can be one of the following symbols:

  before-kick-off
  kick-off (my | enemy)
  kick-in (my | enemy)
  corner-kick (my | enemy)
  goal-kick (my | enemy)
  free-kick (my | enemy)
  play-on
  half_time
  time_up
  extend
  drop-ball
  foul (my | enemy) num
  goal (my | enemy) point"
  (agent-kb-last-referee-msg kb))

(defun last-see-ball(kb)
  "Return the last info avaliable in the KB about the ball."
  (if (agent-kb-last-ball-dir kb)
      (agent-kb-ball-pos kb)
    (let ((ball (agent-kb-ball-pos kb)))
      (when (and ball (< (- (get-time kb) (obj-info-time ball)) 
			 *max-ball-age*))
	ball))))

(defun dir-x-y (kb x y)
  "calculates the direction relative to the player of an object"
  (let ((w (- x (pos-x (agent-kb-player-pos kb))))
	(h (- y (pos-y (agent-kb-player-pos kb))))
	(theta (pos-dir (agent-kb-player-pos kb)))
	alfa)
    (setf alfa
	  (if (zerop w)
	      (if (> h 0) 90 -90)
	    (if (> w 0)
		(deg-atan (/ h w))
	      (if (> h 0)
		  (+ 180 (deg-atan (/ h w)))
		(+ -180 (deg-atan (/ h w)))))))
    (- alfa theta)))

(defun dir-ball (kb)
  (if (agent-kb-last-ball-dir kb)
      (agent-kb-last-ball-dir kb)
    (let ((ball (last-see-ball kb)))
      (if ball
	  (dir-x-y kb (obj-info-x ball) (obj-info-y ball))
	0))))


(defun distance (kb x y)
  "calculates the distance relative to the player of an objcect"
  (sqrt (+ (sqr (- x (pos-x (agent-kb-player-pos kb))))
	   (sqr (- y (pos-y (agent-kb-player-pos kb)))))))

(defun gen-pos-x (kb x w)
  "Calculates a position given a start number and an offset expressed in 16th"
  (let ((rx (remap-x x))
	(rw (remap-w w))
	(ball (last-see-ball kb)))
    (if ball
	(if (> (obj-info-x ball) (pos-x (agent-kb-player-pos kb)))
	    (max (obj-info-x ball) (+ rx rw))
	  (min (obj-info-x ball) rx))
      (+ rx (random rw)))))

(defun gen-pos-y (kb y h)
  "Calculates a position given a start number and an offset expressed in 16th"
  (let ((ry (remap-y y))
	(rh (remap-h h))
	(ball (last-see-ball kb)))
    (if ball
	(if (> (obj-info-y ball) (pos-y (agent-kb-player-pos kb)))
	    (max (obj-info-y ball) (+ ry rh))
	  (min (obj-info-y ball) rh))
      (+ ry (random rh)))))

(defun enemy (kb num)
  (let ((pl (svref (agent-kb-enemy-players kb) num)))
    (if (and pl (< (- (cadr (agent-kb-last-percept kb)) (obj-info-time pl)) 
		   *max-player-age*))
	pl
      nil)))

(defun friend (kb num)
  (let ((pl (svref (agent-kb-team-players kb) num)))
    (if (and pl (< (- (cadr (agent-kb-last-percept kb)) (obj-info-time pl)) 
		   *max-player-age*))
	pl
      nil)))

(defun enemies-near-pos (kb x y)
  (let ((pl nil)
	(count 0))
    (dotimes (i 11 count)
	     (setf pl (enemy kb (+ i 1)))
	     ;; 64 is 8*8 ;)
	     (when (and pl (< (sqr-radius x y
					  (obj-info-x pl)
					  (obj-info-y pl)) 64))
	       (incf count)))))

(defun enemies-near-ball (kb)
  (let ((ball (last-see-ball kb)))
    (if ball
	(enemies-near-pos kb (obj-info-x ball) (obj-info-y ball))
      0)))

(defun friends-near-pos (kb x y)
  (let ((pl nil)
	(count 0))
    (dotimes (i 11 count)
	     (setf pl (friend kb (+ i 1)))
	     ;; 64 is 8*8 ;)
	     (when (and pl (< (sqr-radius x y
					  (obj-info-x pl)
					  (obj-info-y pl)) 64))
	       (incf count)))))

(defun friends-near-ball (kb)
  (let ((ball (last-see-ball kb)))
    (if ball
	(friends-near-pos kb (obj-info-x ball) (obj-info-y ball))
      0))) 

(defun players-near-ball (kb)
  (+ (enemies-near-ball kb) (friends-near-ball kb)))

(defun pass-candidates (kb)
  (let ((ret ())
	(pl nil))
    (dotimes (i 11 ret)
	     (setf pl (friend kb (+ i 1)))
	     (when (and pl (= (obj-info-time pl) (get-time kb)))
	       (setf ret (append ret (list (cons (+ i 1) pl))))))))

(defun ok-to-pass (kb el)
  (let* ((emp (svref *pass-empathy* (car el)))
	(obj (cdr el))
	(dist (distance kb (obj-info-x obj) (obj-info-y obj)))
	(semi (- (obj-info-x obj) (pos-x (agent-kb-player-pos kb))))
	(ret 0))
    ;; 1/1+x is the function
    (incf ret (* (/ 1 (+ 1 (abs (- *pass-ideal-distance* dist)))) emp))
    (incf ret (/ (* semi (- half-pitch-length (pos-x (agent-kb-player-pos kb))))
		 (* 4 half-pitch-length half-pitch-length)))
    ret))

(defun no-pass-ball (kb el)
  (let* ((obj (cdr el))
	 (x (obj-info-x obj))
	 (y (obj-info-y obj))
	 (dir (+ (obj-info-dir obj) (dir-x-y kb x y)))
	 (pl nil)
	 (count 0))
    (dotimes (i 11)
	     (setf pl (friend kb (+ i 1)))
	     (when (and 
		    pl 
		    (< (abs (- 
			     (+ (dir-x-y kb (obj-info-x pl) (obj-info-y pl))
				(obj-info-dir pl))
			     dir)) 10))
	       (incf count)))
    (/ count 11)))

(defun select-candidate (kb cand)
  (if cand
      (let ((ret nil)
	    (max -1)
	    (pos (agent-kb-player-pos kb))
	    val
	    my-val)
	(dolist 
	 (el cand ret)
	 (setf val (- (* *alpha-pass* (ok-to-pass kb el)) 
		      (* *beta-no-pass* (no-pass-ball kb el))))
	 (when (> val max)
	   (setf max val
		 ret (cdr el))))

	(setf my-val (- 
		      (/ 10 (* 2 half-pitch-length)) ; 10 meter if I keep ball
		      (/ (enemies-near-pos kb (pos-x pos) (pos-y pos)) 
			 11)))
;(format t "Max is ~A and my-val is ~A~%" max my-val)

	(when (and (>= max my-val)
		   (not (offside? kb ret)))
	  ret))
    nil))

(defun ball-distance (kb)
  (let ((ball (last-see-ball kb)))
    (when ball
	  (distance kb (obj-info-x ball) (obj-info-y ball)))))
