;;;;
;;;; Last update: 08/08/1999
;;;; @author Antonio Cisternino
;;;; @version 0.3
;;;;
;;;; Definition of the class that defines a player
;;;;
;;;; Changes:
;;;;
;;;; 08/08/1999: Now the player is implemented using LISP functions rather
;;;;             than CLOS.
;;;;
;;;; 12/3/1999: Now the read-last-percept function is used
;;;;
;;;; 2/1/1999: Added a lot of documentation
;;;;
;;;; 1/1/1999: The layers are now loaded from the plans directory
;;;;
;;;; 1/1/1999: The team name has changed name

(defstruct ialp-player
  side
  number
  team
  state
  percept
  planner
  goalie
  (kb (make-agent-kb)))

(defun make-player (type team)
  "Create a ialp player"
  (make-ialp-player :goalie (eq type 'goalie) :team team))

(defun connect (player host)
  "Connect the player to the host"
  (connect-to-server host)
  (do-action
   (if (ialp-player-goalie player)
       `(init ,(ialp-player-team player) (goalie) (version 5.00))
     `(init ,(ialp-player-team player) (version 5.00))))
  (let ((percept (read-percept)))
    (if (eq (car percept) 'error)
	(eval percept)
      (setf 
       (ialp-player-side player) (second percept)
       (ialp-player-number player) (third percept)
       (ialp-player-state player) (fourth percept)
       (ialp-player-percept player) percept))
    (rc-set-jersey-num (third percept)))
  player)

(defun reconnect (player host)
  "Reconnect the player to the host"
  (do-action (list 'init (ialp-player-team player) (ialp-player-number player)))
    (let ((percept (read-percept)))
      (if (eq (car percept) 'error)
	  (error "can_not_reconnect")
	(setf 
	 (ialp-player-side player) (second percept)
	 (ialp-player-state player) (third percept)
	 (ialp-player-percept player) percept)))
    player)

(defun get-percept (p)
  "Get the last percept sent by the server and set the percept attribute
   of the ialp-player"
  (setf (ialp-player-percept p) (read-last-percept)))

(defun pos-agent (x y)
  "Set the agent position"
  (do-action `(move ,x ,y)))

(defun run-agent (p)
  "Loop of the agent"
  (init-agent p)
  (loop
   (get-percept p)
   (let ((action (plan-action p (ialp-player-percept p))))
      (when action
	(if (listp (car action))
	    (do-actions action)
	  (do-action action))))))

(defun plan-action (p percept)
  "This method defines a step of simulation"

  ;; Reset the turn trashing
  (rc-turn-trash-on)

  ;; Update the KB
  (tell-KB (ialp-player-kb p) percept)

  (when (not (eq (car percept) 'sense_body))
    ;; Change the plan if necessary
    (update-plans p)
    
#| Is this mechanism useful?
    ;; Unless there is an instinctive action make a step of the
    ;; current plan
    (let ((action (immediate-action p percept)))
      (unless action
	(setf action (make-step p)))
      action)))
|#

    (make-step p)))

(defun init-agent (p)
  "Initialize the agent by loading the various plans. This method
   searches for the file '*ialp-dir-plans*/layers-<n>.lst' where
   *ialp-dir-plans* is the directory which contains the plans for
   the various players and <n> is the number of the player assigned
   by the server.
   The file layers-<n>.lst contains a list of names. Each name refers
   to a specific layer of the planner."

  ;; Tells to the KB the team name and the initial side of the team
  (setf (agent-kb-team (ialp-player-kb p)) (ialp-player-team p))
  (setf (agent-kb-side (ialp-player-kb p)) (if (eq 'l (ialp-player-side p)) #\L #\R))

  ;; Build the layers chain of the planner reading the layers-<n>.lst
  ;; for first and the various files listed later.
  (let ((lay nil) l)
    (with-open-file (in (format nil "~A/layers-~A.lst" *ialp-dir-plans* 
				(ialp-player-number p)) :direction :input)
		    (setf l (read-line in nil nil))
		    (format t "Loading the top-layer named '~A'...~%" l)
		    (ialp-load-plan l)
		    (do ((el (read-line in nil nil) (read-line in nil nil)))
			((not el))
			(format t "Loading layer ~A...~%" el)
			(setf l (load-layer el))
			(setf (layer-player l) p)
			(setf (layer-player-kb l) (ialp-player-kb p))
			(setf (layer-superior-layer l) lay)
			(if lay (setf (layer-inferior-layer lay) l))
			(setf lay l)))
    (setf (ialp-player-planner p) lay))

  ;; Move the player to the initial position
  (do-action `(move ,*start-x* ,*start-y*))

  ;; Wait for a second
  (sleep 1)

  ;; Turn the player to the start direction
  (do-action `(turn ,(if (eq (agent-kb-side (ialp-player-kb p)) #\L) 
			 *start-dir*
		       (if (>= *start-dir* 0)
			   (- 180 *start-dir*)
			 (- -180 *start-dir*)))))

  ;; Tell to the KB the starting position
  (setf (agent-kb-player-pos (ialp-player-kb p)) (make-pos :x *start-x* :y *start-y* 
					     :dir *start-dir*)))

(defun immediate-action (p percept)
  "Decide if it is possible to make an immediate action excluding
   the planner.
   For the moment is disabled"
  nil)

(defun update-plans (p)
  "Update the plan: if the environment is changed in some way probably
   we must rebuild the plan"
  (update-plan (ialp-player-planner p)))

(defun make-step (p)
  "Returns an action due to a step of execution of the current plan."
  (make-layer-step (ialp-player-planner p)))
