;;;; layer.lsp
;;;; 
;;;; Last update: 10/08/1999
;;;; @author Antonio Cisternino
;;;; @version 0.4
;;;;
;;;; Purpose: this file defines a layer of the hierarchical planner.
;;;;
;;;; Changes:
;;;;
;;;; 10/08/1999: Now the layer is implemented as struct+fn rather
;;;;             than using CLOS.
;;;;
;;;; 6/1/1999: Fixed a bug in method 'make-layer-step'. Sometimes a (nil)
;;;;           was fired instead of a command.
;;;;
;;;; 5/1/1999: Added the *while* support in the plan definitions.
;;;;           *while* operations cannot be nested and the
;;;;           general syntax for a while is:
;;;;           (... (*while* <guard> <plan>) ...)
;;;;           Where <guard> is an expression and plan is a list
;;;;           of actions.
;;;;
;;;; 3/1/1999: Changed the function update-plan
;;;;
;;;; 2/1/1999: Added a lot of documentation
;;;;
;;;; 1/1/1999: Changed the location from which a layer is load

;;; Defines a layer of the planner
(defstruct layer
  ;; Player that holds this layer of the planner
  player

  ;; KB of that agent
  player-kb

  ;; Hashtable that contains all the plans of this layer
  plans

  ;; Upper layer (if nil is the top layer)
  superior-layer

  ;; Lower layer (if nil is the bottom layer)
  inferior-layer

  ;; Current execution plan
  current-plan

  ;; While guard. If is nil there is not a while active.
  while-guard

  ;; While body. Is the body of the while
  while-body

  ;; While body temp. Used for a complex plan like the current attribute.
  while-current-plan

  ;; Function that evauates when it is necessary to replan
  ;; This function must return 'UP, 'DOWN or nil to say which
  ;; layers must replan.
  update)

(defun abort (l dir)
  "Aborts the current executing plan for the layer.
   I the second parameter is UP propagates the message up through
   the layers chain. If he is 'DOWN propagates the message to
   inferior layers."
  (setf (layer-current-plan l) nil
	(layer-while-guard l) nil
	(layer-while-body l) nil
	(layer-while-current-plan l) nil)
  (if (eq dir 'UP)
      (when (layer-superior-layer l) (abort (layer-superior-layer l) 'UP))
    (if (eq dir 'DOWN)
	(when (layer-inferior-layer l) (abort (layer-inferior-layer l) 'DOWN)))))

(defun update-plan (l)
  "Signals a perception received. The layer can ask the KB for
   important informations and decide to replan."
  (let ((ret (apply (layer-update l) (list (layer-player l)))))
    (when ret (abort l ret))

  ;; If I've aborted all plans of superior layers it is not
  ;; necessary to update
  (when (and (not (eql ret 'UP)) 
	     (layer-superior-layer l)) 
    (update-plan (layer-superior-layer l)))))

(defun load-layer (layer-name)
  "Creates and loads a layer from file."
  (let ((ret (make-layer)))
    (load-layer-from-file ret layer-name)
    ret))

(defun load-layer-from-file (l file)
  "Loads a layer from a file"
  (ialp-load-plan file)
  (setf (layer-plans l)
	(eval (read-from-string (format nil "(new-~A)" file)))
	(layer-update l)
	(eval (read-from-string (format nil "#'update-~A" file)))))

(defmacro **make-step (pl)
  `(let ((ret (car ,pl)))
    (setf ,pl (cdr ,pl))
    (call l (car ret) (cdr ret))))

(defun make-layer-step (l)
  "Makes a layer step"

  ;; Cannot find a plan for the layer
  ;; Ask to superior layer, if exists, or to the top-level-decision function
  (unless (or (layer-while-guard l) (layer-current-plan l))
    (setf (layer-current-plan l)
	  (if (layer-superior-layer l)
	      (make-layer-step (layer-superior-layer l))
	    (top-level-decision (layer-player l)))))
  
  ;; Is in execution a While?
  (if (layer-while-guard l)
      ;; The body of while must be executed?
      (if (layer-while-current-plan l)

	  ;; Make a step of execution for the body
	  (**make-step (layer-while-current-plan l))

	;; No body in execution: evaluates the guard
	(if (eval (layer-while-guard l))
	    (progn
	      ;; The guard is ok: initialize the body plan
	      (setf (layer-while-current-plan l) (layer-while-body l))

	      ;; Execute the first step of the body
	      (**make-step (layer-while-current-plan l)))

	  ;; The guard is false
	  (progn
	    ;; Reset the while status
	    (setf (layer-while-guard l) nil
		  (layer-while-current-plan l) nil)
	    
	    ;; Call recursively to execute the first step of the While
	    (make-layer-step l))))

    ;; No while in execution, next action is a while?
    (if (eql (caar (layer-current-plan l)) '*while*)
	(progn
	  ;; Initialize the while
	  (setf (layer-while-guard l) (second (car (layer-current-plan l)))
		(layer-while-body l) (third (car (layer-current-plan l))))

	  ;; Eat the action from the current plan
	  (setf (layer-current-plan l) (cdr (layer-current-plan l)))

	  ;; Call recursively to execute the first step of the While
	  (make-layer-step l))

      ;; No while, executes as usually the action
      (**make-step (layer-current-plan l)))))
	    
#|
  ;; Evaluates a step of the plan
  (let ((ret (car (current-plan l))))
    (setf (current-plan l) (cdr (current-plan l)))
    (call l (car ret) (cdr ret))))
|#
    
(defun call (l fun args)
  "Calls an action. An action must return a list of commands
   expressed in the inferior layer defined language.
   The usual list returned bust be a list where each element
   is a cons where the car contains the name of the action
   and the cadr contains a list of arguments that must be
   passed to the function.
   If the action is not understanded by this layer is directly
   reflected to the inferior layer."
  (let ((el (gethash fun (layer-plans l))) fn)
    (if el
	(progn
	  (setf fn (if (> (car el) 1) ; Ottimizza se ho solo una funzione 
					; non scelgo
		       (cdr (nth (truncate (random (car el))) (cdr el)))
		     (cdadr el)))
	  (apply fn l args))
      `((,fun . ,args)))))

(defun list-commands (l)
  "Print all commands defined in this layer"
  (maphash #'(lambda (key el)
	       (format t "- ~A~%" (princ-to-string key))
	       (dolist (fun (cdr el))
		       (format t "~A~%"
			       (documentation (car fun) 'function)))
	       (format t "~%")) (layer-plans l)))

