;;;; 
;;;; Last update: 9/1/1999
;;;; @author Antonio Cisternino
;;;; @version 0.1
;;;;
;;;; Purpose: here there are the macros available
;;;;
;;;; Changes:
;;;;
;;;; 25/2/1999: Added the macros remap-x and remap-y
;;;;
;;;; 16/1/1999: Added the section of the language for defining plans
;;;;
;;;; 14/1/1999: Added the macro deg-atan

;;; Calc macros: these macros are used to enhance the math support
;;; and handle with degrees in sin and cos function.
(defmacro deg-to-rad (arg)
  "Returns the radians given the degrees value."
  `(* (/ ,arg 180) 3.141))

(defmacro deg-atan (arg)
  "Returns the atan in degrees rather than radians."
  `(* (/ (atan ,arg) 3.141) 180))

(defmacro deg-sin (arg)
  "Returns the sin given an argument expressed in degrees rather than
   radians."
  `(sin (* (/ ,arg 180) 3.141)))

(defmacro deg-cos (arg)
  "Returns the cos given an argument expressed in degrees rather than
   radians."
  `(cos (* (/ ,arg 180) 3.141)))

(defmacro sqr (x)
  "Calculate the square of a given number."
  `(* ,x ,x))

(defmacro sqr-dist (x1 y1 x2 y2)
  "Calculate the square of the distance. Util for ordering purposes..."
  `(+ ,(sqr (- x1 x2)) ,(sqr (- y1 y2))))

(defmacro rc-distance (x1 y1 x2 y2)
  "Calculate the euclidean distance between two points."
  `(sqrt (+ (* (- ,x1 ,x2) (- ,x1 ,x2)) (* (- ,y1 ,y2) (- ,y1 ,y2)))))

;;; Following macros are intended to express qualitative quantities
;;; to allow the planning to be more qualitative rather than quantitative.
;;; Most of these macros defines a range and select a value within the
;;; the range using a uniform distribution.
(defmacro sprint ()
  '100)

(defmacro very-fast ()
  `(+ 80 (random 21)))

(defmacro fast ()
  `(+ 70 (random 11)))

(defmacro light-fast ()
  `(+ 60 (random 15)))

(defmacro normal ()
  `(+ 50 (random 11)))

(defmacro normal-slow ()
  `(+ 45 (random 6)))

(defmacro slow ()
  `(+ 30 (random 6)))

(defmacro very-slow ()
  `(+ 20 (random 11)))

(defmacro walk ()
  `(+ 5 (random 16)))

(defmacro back ()
  `(- -5 (random 16)))

(defmacro fast-back ()
  `(- -20 (random 11)))

(defmacro obj-near (kb obj)
  `(let* ((ppos (agent-kb-player-pos ,kb))
	 (x1 (pos-x ppos))
	 (y1 (pos-y ppos))
	 (x2 (obj-info-x ,obj))
	 (y2 (obj-info-y ,obj)))
    (if (< (rc-distance x1 y1 x2 y2) 10)
	 t
       nil)))

(defmacro obj-very-near (kb obj)
  `(let* ((ppos (agent-kb-player-pos ,kb))
	 (x1 (pos-x ppos))
	 (y1 (pos-y ppos))
	 (x2 (obj-info-x ,obj))
	 (y2 (obj-info-y ,obj)))
    (if (< (rc-distance x1 y1 x2 y2) 5)
	 t
       nil)))

(defmacro remap-x (x)
  `(- (* ,x ,*nth-pitch-length*) ,half-pitch-length))

(defmacro remap-w (w)
  `(* ,w ,*nth-pitch-length*))

(defmacro remap-y (y)
  `(- (* ,y ,*nth-pitch-width*) ,half-pitch-width))

(defmacro remap-h (h)
  `(* ,h ,*nth-pitch-width*))

(defmacro sqr-radius (x1 y1 x2 y2)
  `(let ((dx (- ,x1 ,x2))
	 (dy (- ,y1 ,y2)))
     (+ (* dx dx) (* dy dy))))

(defmacro get-time (kb)
  `(cadr (agent-kb-last-percept ,kb)))

;;;; Language definition

;;; This variable holds the file name used for generate
;;; the names when loading plans.
(defvar *plan-file-name*)

(defun plan-file (name)
  "Used to declare the file name."
  (setf *plan-file-name* name))

(defun plan-name (name)
  "Calculate the name of the plan file as '*plan-file-name*-layer' and
   prepend the prefix 'name-'."
  (read-from-string (format nil "~A-~A-layer" name *plan-file-name*)))

(defun action-name (name)
  "Calculate the name of an action as '*plan-file-name*-name'."
  (format nil "~A-~A" *plan-file-name* (princ-to-string name)))

(defmacro def-empty-update ()
  "Generate an empty update function for a plan file. For example, the
   macro expanded in the file player-base-layer.lsp returns:

   (defun update-player-base-layer (ag) nil)"
  `(defun ,(plan-name "update") (ag) nil))

(defmacro defupdate (&rest args)
  "Generate the skeleton of the update function for a plan file. All the
   arguments passed to the macro are bypassed to the defun form. For
   example in the basic-layer.lsp file returns:

   (defun update-basic-layer  ...)

   Note that, because is a callback of the interpreter of plans, the
   list of variables *cannot* be specified and the variable 'ag' holds
   a reference to the agent object that invokes the function. The
   variable 'kb' holds a referece to the kb of that agent."
  `(defun ,(plan-name "update") (ag)
     (let ((kb (ialp-player-kb ag)))
     ,@args)))

(defmacro defactions (&rest args)
  "This macro expands to a form which declares wich actions are related
   to a layer. The syntax allowed is:

  (defactions
    {(action-list ...)}+)

  The macro expands to (in basic-layer.lsp):

  (defun new-basic-layer ()
     (let ((ret (make-hash-table)))
       ... Code generated by the (action-list ...) macros
       ret))

  To force the expansion of the various (action-list) I use the
  macroexpand-1."
  `(defun ,(plan-name "new") ()
     (let ((ret (make-hash-table)))
       ;; Expands all the forms (action-list)
       ,@(let ((retl ()))
	   (dolist
	    (el args)
	    (when (not (and (consp el) (eq (car el) 'action-list)))
	      (error "defactions must contains only action-list forms!"))
	    (setf retl (append
		       retl
		       (list (macroexpand-1 el)))))
	   retl) ;; End of the evaluated form
       ret)))

(defmacro action-list (name &rest fn)
  "Register a list of actions under a give name. The macro
   *must* be used in the (defactions ...) macro.
   The usage is:

   (action-list action-name {action-impl-name}*)

   The action-name defines the name that can be used by the upper
   layers to invoke an action. The action-impl-name are the names
   passed to the (defaction ...) form to implement the action.
   If more implementation are specified the interpreter selects
   one from the list of the equivalent implementations using an
   appropriate policy. If no implementation name is given the
   form assumes that there is only one implementation for the
   action and the name is the same of the action.
   The form (action-list turn) is equivalent to the form
   (action-list turn turn) and expands to (in the file basic-layer.lsp):

   (setf (gethash 'turn ret)
         (cons 1 (list (cons 'basic-turn #'basic-turn))))

   If the form is (action-list turn turn1 turn2) expands to:

   (setf (gethash 'turn ret)
         (cons 2 (list (cons 'basic-turn1 #'basic-turn1)
                       (cons 'basic-turn2 #'basic-turn2))))

   and so on..."   

  ;; If there is only one argument the name is assumed
  (let ((fnl  (if (zerop (length fn))
		  `(,name)
		fn)))
    `(setf
      (gethash (quote ,name) ret)
      (cons ,(length fnl)
	    (list
	     ;; Calculates the list of implementations names
	     ,@(let ((ret ()))
		 (dolist 
		  (el fnl)
		  (setf ret (append
			     ret
			     (let ((n (read-from-string (action-name el))))
			       `((cons (quote ,n)
				       (function ,n)))))))
		 ret))))))

(defmacro defaction (name bind &rest args)
  "Defines an action. This expands to a defun with a standard prototype
   (required by the interpreter) and bind the parameters to the real
   parameters list.
   The variable 'l' contains the reference to the layer that holds the
   implementation of the function. The 'kb' variable contains the reference
   to the agent KB. The 'args' variable contains the parameters passed
   to the function without the binding.
   The (defaction turn (dir) 'Doc' `(turn ,dir)) expands to:

   (defun basic-turn (l &args)
     'Doc'
     (let ((kb (player-kb l))
           (dir (nth 0 args)))
       `(turn ,dir)))
   ."

  (let ((doc ""))
    (when (stringp (car args))
      (setf doc (car args))
      (setf args (cdr args)))
    `(defun ,(read-from-string (action-name name)) (l &rest args)
       ,doc
       (let ((kb (layer-player-kb l))
	     ;; Preparates the binding list
	     ,@(let ((pos 0)
		     (ret ()))
		 (dolist
		  (el bind)
		  (setf ret
			(append
			 ret
			 `((,el (nth ,pos args)))))
		  (incf pos))
		 ret)) ;; Concludes the evaluated parameter
	 ,@args))))
